import numpy
import pandas
import pandas.testing as pdtest
import pytest
import seaborn
from matplotlib import pyplot

from wqio import hydro
from wqio.tests import helpers

BASELINE_IMAGES = "_baseline_images/hydro_tests"
TOLERANCE = helpers.get_img_tolerance()
FIVEMIN = pandas.offsets.Minute(5)


class fakeStormSublcass(hydro.Storm):
    is_subclassed = True


def hr_simple():
    storm_file = helpers.test_data_path("teststorm_simple.csv")
    orig_record = (
        pandas.read_csv(storm_file, index_col="date", parse_dates=True)
        .resample(FIVEMIN)
        .asfreq()
        .fillna(0)
    )

    return hydro.HydroRecord(
        orig_record,
        precipcol="rain",
        inflowcol="influent",
        outflowcol="effluent",
        outputfreqMinutes=5,
        minprecip=1.5,
        intereventHours=3,
    )


@pytest.fixture()
def hr_simple_fixture():
    return hr_simple()


def hr_singular():
    storm_file = helpers.test_data_path("teststorm_singular.csv")
    orig_record = (
        pandas.read_csv(storm_file, index_col="date", parse_dates=True)
        .resample(FIVEMIN)
        .asfreq()
        .fillna(0)
    )
    return hydro.HydroRecord(
        orig_record,
        precipcol="rain",
        inflowcol="influent",
        outflowcol=None,
        outputfreqMinutes=5,
        intereventHours=3,
    )


def hr_first_obs():
    storm_file = helpers.test_data_path("teststorm_firstobs.csv")
    orig_record = (
        pandas.read_csv(storm_file, index_col="date", parse_dates=True)
        .resample(FIVEMIN)
        .asfreq()
        .fillna(0)
    )
    return hydro.HydroRecord(
        orig_record,
        precipcol="rain",
        inflowcol="influent",
        outflowcol=None,
        outputfreqMinutes=5,
        intereventHours=3,
    )


def hr_diff_storm_class():
    storm_file = helpers.test_data_path("teststorm_simple.csv")
    orig_record = (
        pandas.read_csv(storm_file, index_col="date", parse_dates=True)
        .resample(FIVEMIN)
        .asfreq()
        .fillna(0)
    )
    return hydro.HydroRecord(
        orig_record,
        precipcol="rain",
        inflowcol="influent",
        outflowcol=None,
        outputfreqMinutes=5,
        minprecip=1.5,
        stormclass=fakeStormSublcass,
        intereventHours=3,
    )


def setup_storms(filename, baseflow=None):
    storm_file = helpers.test_data_path(filename)
    orig_record = (
        pandas.read_csv(storm_file, index_col="date", parse_dates=True)
        .resample(FIVEMIN)
        .asfreq()
        .fillna(0)
    )

    baseflowcol = None
    if baseflow is not None:
        orig_record = orig_record.assign(baseflow=orig_record["influent"] < baseflow)
        baseflowcol = "baseflow"

    parsed = hydro.parse_storm_events(
        orig_record,
        3,
        5,
        precipcol="rain",
        inflowcol="influent",
        baseflowcol=baseflowcol,
        debug=True,
    )
    return parsed


@pytest.mark.parametrize(("wetcol", "expected"), [("A", 1), ("B", 0)])
def test_wet_first_row(wetcol, expected):
    df = pandas.DataFrame(
        {"A": [True, True, False], "B": [False, False, True], "Z": [0, 0, 0]}
    ).pipe(hydro._wet_first_row, wetcol, "Z")
    assert df.iloc[0].loc["Z"] == expected


def test_wet_window_diff():
    is_wet = pandas.Series([0, 1, 1, 1, 0, 0, 1, 0, 0, 0])
    result = hydro._wet_window_diff(is_wet, 3)
    expected = pandas.Series([numpy.nan, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, -1.0])
    pdtest.assert_series_equal(result, expected)


@pytest.mark.parametrize(
    ("stormfile", "baseflow"),
    [
        ("teststorm_simple.csv", None),
        ("teststorm_firstobs.csv", None),
        ("teststorm_singular.csv", None),
        ("teststorm_simple.csv", 2),
    ],
)
def test_number_of_storms(stormfile, baseflow):
    parsed = setup_storms(stormfile, baseflow=baseflow)
    last_storm = parsed["storm"].max()
    assert last_storm == 5


@pytest.mark.parametrize(
    ("stormfile", "baseflow", "sn", "idx", "known"),
    [
        ("teststorm_simple.csv", None, 1, 0, "2013-05-18 13:40:00"),
        ("teststorm_simple.csv", None, 1, -1, "2013-05-19 01:45:00"),
        ("teststorm_simple.csv", None, 2, 0, "2013-05-19 06:10"),
        ("teststorm_simple.csv", None, 2, -1, "2013-05-19 11:35"),
        ("teststorm_simple.csv", 2, 1, 0, "2013-05-18 14:05:00"),
        ("teststorm_simple.csv", 2, 1, -1, "2013-05-19 01:40:00"),
        ("teststorm_simple.csv", 2, 2, 0, "2013-05-19 06:35"),
        ("teststorm_simple.csv", 2, 2, -1, "2013-05-19 11:30"),
        ("teststorm_singular.csv", None, 1, 0, "2013-05-18 13:40:00"),
        ("teststorm_singular.csv", None, 1, -1, "2013-05-19 01:45:00"),
        ("teststorm_singular.csv", None, 2, 0, "2013-05-19 07:00"),
        ("teststorm_singular.csv", None, 2, -1, "2013-05-19 07:05"),
        ("teststorm_firstobs.csv", None, 1, 0, "2013-05-18 12:05"),
        ("teststorm_firstobs.csv", None, 1, -1, "2013-05-19 01:45:00"),
        ("teststorm_firstobs.csv", None, 2, 0, "2013-05-19 06:10"),
        ("teststorm_firstobs.csv", None, 2, -1, "2013-05-19 11:35"),
    ],
)
def test_storm_start_end(stormfile, baseflow, sn, idx, known):
    parsed = setup_storms(stormfile, baseflow=baseflow)
    storm = parsed[parsed["storm"] == sn]
    ts = pandas.Timestamp(known)
    assert storm.index[idx].strftime("%X %x") == ts.strftime("%X %x")


@pytest.mark.xfail(TOLERANCE > 15, reason="GH Action weirdness")
@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_HydroRecord_histogram():
    stormfile = helpers.test_data_path("teststorm_simple.csv")
    orig_record = (
        pandas.read_csv(stormfile, index_col="date", parse_dates=True)
        .resample(FIVEMIN)
        .asfreq()
        .fillna(0)
    )
    hr = hydro.HydroRecord(
        orig_record,
        precipcol="rain",
        inflowcol="influent",
        outflowcol=None,
        outputfreqMinutes=5,
        minprecip=1.5,
        intereventHours=3,
    )

    with seaborn.axes_style("ticks"):
        fig = hr.histogram("Total Precip Depth", [4, 6, 8, 10])
    return fig.figure


def test_HydroRecord_attr(hr_simple_fixture):
    expected_std_columns = ["baseflow", "rain", "influent", "effluent", "storm"]
    assert isinstance(hr_simple_fixture._raw_data, pandas.DataFrame)

    assert isinstance(hr_simple_fixture.data, pandas.DataFrame)
    assert isinstance(hr_simple_fixture.data.index, pandas.DatetimeIndex)
    assert sorted(hr_simple_fixture.data.columns.tolist()) == sorted(expected_std_columns)

    assert isinstance(hr_simple_fixture.all_storms, dict)

    assert isinstance(hr_simple_fixture.storms, dict)

    assert isinstance(hr_simple_fixture.storm_stats, pandas.DataFrame)


@pytest.mark.parametrize(
    ("hr", "expected"),
    [
        (hr_simple(), 5),
        (hr_singular(), 5),
        (hr_first_obs(), 5),
        (hr_diff_storm_class(), 5),
    ],
)
def test_HydroRecord_number_storms(hr, expected):
    assert hr.data["storm"].max() == expected


@pytest.mark.parametrize(
    ("hr", "expected"),
    [
        (hr_simple(), pandas.Timestamp("2013-05-18 13:40:00")),
        (hr_singular(), pandas.Timestamp("2013-05-18 13:40:00")),
        (hr_first_obs(), pandas.Timestamp("2013-05-18 12:05")),
        (hr_diff_storm_class(), pandas.Timestamp("2013-05-18 13:40:00")),
    ],
)
def test_HydroRecord_first_storm_start(hr, expected):
    storm1 = hr.data[hr.data["storm"] == 1]
    assert storm1.index[0] == expected


@pytest.mark.parametrize(
    ("hr", "expected"),
    [
        (hr_simple(), pandas.Timestamp("2013-05-19 01:55:00")),
        (hr_singular(), pandas.Timestamp("2013-05-19 01:45:00")),
        (hr_first_obs(), pandas.Timestamp("2013-05-19 01:45:00")),
        (hr_diff_storm_class(), pandas.Timestamp("2013-05-19 01:45:00")),
    ],
)
def test_HydroRecord_first_storm_end(hr, expected):
    storm1 = hr.data[hr.data["storm"] == 1]
    assert storm1.index[-1] == expected


@pytest.mark.parametrize(
    ("hr", "expected"),
    [
        (hr_simple(), pandas.Timestamp("2013-05-19 06:10")),
        (hr_singular(), pandas.Timestamp("2013-05-19 07:00")),
        (hr_first_obs(), pandas.Timestamp("2013-05-19 06:10")),
        (hr_diff_storm_class(), pandas.Timestamp("2013-05-19 06:10")),
    ],
)
def test_HydroRecord_second_storm_start(hr, expected):
    storm2 = hr.data[hr.data["storm"] == 2]
    assert storm2.index[0] == expected


@pytest.mark.parametrize(
    ("hr", "expected"),
    [
        (hr_simple(), pandas.Timestamp("2013-05-19 11:55")),
        (hr_singular(), pandas.Timestamp("2013-05-19 07:05")),
        (hr_first_obs(), pandas.Timestamp("2013-05-19 11:35")),
        (hr_diff_storm_class(), pandas.Timestamp("2013-05-19 11:35")),
    ],
)
def test_HydroRecord_second_storm_end(hr, expected):
    storm2 = hr.data[hr.data["storm"] == 2]
    assert storm2.index[-1] == expected


def test_HydroRecord_interevent_stuff(hr_simple_fixture):
    assert hr_simple_fixture.intereventPeriods == 36
    assert hr_simple_fixture.intereventHours == 3


def test_HydroRecord_storm_stat_columns(hr_simple_fixture):
    expected = [
        "Storm Number",
        "Antecedent Days",
        "Season",
        "Start Date",
        "End Date",
        "Duration Hours",
        "Peak Precip Intensity",
        "Total Precip Depth",
        "Total Inflow Volume",
        "Peak Inflow",
        "Total Outflow Volume",
        "Peak Outflow",
        "Peak Lag Hours",
        "Centroid Lag Hours",
    ]
    assert hr_simple_fixture.storm_stats.columns.tolist() == expected


def test_HydroRecord_storm_attr_lengths(hr_simple_fixture):
    assert len(hr_simple_fixture.storms) == 2
    assert len(hr_simple_fixture.all_storms) == 5


def test_HydroRecord_storm_stats(hr_simple_fixture):
    cols = [
        "Storm Number",
        "Antecedent Days",
        "Peak Precip Intensity",
        "Total Precip Depth",
    ]

    expected = pandas.DataFrame(
        {
            "Storm Number": [1, 5],
            "Antecedent Days": [numpy.nan, 0.701389],
            "Peak Precip Intensity": [1.200, 1.200],
            "Total Precip Depth": [2.76, 4.14],
        }
    )
    pdtest.assert_frame_equal(hr_simple_fixture.storm_stats[cols], expected[cols])


@pytest.mark.parametrize(
    ("value", "error"),
    [
        (pandas.Timestamp("2013-05-19 11:11"), None),
        (str(pandas.Timestamp("2013-05-19 11:11")), None),
        (pandas.Timestamp("2013-05-19 11:11").to_pydatetime(), None),
        ("junk", ValueError),
    ],
)
@pytest.mark.parametrize("smallstorms", [True, False])
def test_getStormFromTimestamp_(hr_simple_fixture, value, smallstorms, error):
    with helpers.raises(error):
        sn, storm = hr_simple_fixture.getStormFromTimestamp(value, smallstorms=smallstorms)
        assert sn == 2
        if not smallstorms:
            assert storm is None
        else:
            assert isinstance(storm, hydro.Storm)


def test_getStormFromTimestame_big(hr_simple_fixture):
    ts = pandas.Timestamp("2013-05-18 13:40:00")
    sn, storm = hr_simple_fixture.getStormFromTimestamp(ts)
    assert sn == 1
    assert isinstance(storm, hydro.Storm)


@pytest.mark.parametrize(
    ("lbh", "expected", "error"),
    [(6, 2, None), (2, None, None), (-3, None, ValueError)],
)
def test_getStormFromTimestamp_lookback(hr_simple_fixture, lbh, expected, error):
    with helpers.raises(error):
        ts = pandas.Timestamp("2013-05-19 14:42")
        sn, storm = hr_simple_fixture.getStormFromTimestamp(ts, lookback_hours=lbh)
        if expected:
            assert sn == expected
        else:
            assert sn is None


def test_HydroRecord_subclassed_storms():
    hr = hr_diff_storm_class()
    assert all([hr.all_storms[sn].is_subclassed for sn in hr.all_storms])


@pytest.fixture
def basic_storm():
    hr = hr_simple()
    storm = hydro.Storm(
        hr.data,
        2,
        precipcol=hr.precipcol,
        inflowcol=hr.inflowcol,
        outflowcol=hr.outflowcol,
        freqMinutes=hr.outputfreq.n,
    )
    return storm


def test_Storm_columns(basic_storm):
    expected = ["rain", "influent", "effluent", "baseflow", "storm"]
    assert sorted(basic_storm.data.columns.tolist()) == sorted(expected)


def test_Storm_index_type(basic_storm):
    assert isinstance(basic_storm.data.index, pandas.DatetimeIndex)


def test_Storm_start(basic_storm):
    assert hasattr(basic_storm, "start")
    ts = pandas.Timestamp("2013-05-19 06:10")
    assert basic_storm.start.strftime("%X %x") == ts.strftime("%X %x")


def test_Storm_end(basic_storm):
    assert hasattr(basic_storm, "end")
    ts = pandas.Timestamp("2013-05-19 11:55")
    assert basic_storm.end.strftime("%X %x") == ts.strftime("%X %x")


def test_Storm_season(basic_storm):
    assert hasattr(basic_storm, "season")
    assert basic_storm.season == "spring"


def test_Storm_season_setter(basic_storm):
    basic_storm.season = "fall"
    assert basic_storm.season == "fall"


def test_Storm_precip_start(basic_storm):
    assert hasattr(basic_storm, "precip_start")
    ts = pandas.Timestamp("2013-05-19 06:10")
    assert basic_storm.precip_start.strftime("%X %x") == ts.strftime("%X %x")


def test_Storm_precip_end(basic_storm):
    assert hasattr(basic_storm, "precip_end")
    ts = pandas.Timestamp("2013-05-19 09:10")
    assert basic_storm.precip_end.strftime("%X %x") == ts.strftime("%X %x")


def test_Storm_inflow_start(basic_storm):
    assert hasattr(basic_storm, "inflow_start")
    ts = pandas.Timestamp("2013-05-19 06:25")
    assert basic_storm.inflow_start.strftime("%X %x") == ts.strftime("%X %x")


def test_Storm_inflow_end(basic_storm):
    assert hasattr(basic_storm, "inflow_end")
    ts = pandas.Timestamp("2013-05-19 11:30")
    assert basic_storm.inflow_end.strftime("%X %x") == ts.strftime("%X %x")


def test_Storm_outflow_start(basic_storm):
    assert hasattr(basic_storm, "outflow_start")
    ts = pandas.Timestamp("2013-05-19 06:50")
    assert basic_storm.outflow_start.strftime("%X %x") == ts.strftime("%X %x")


def test_Storm_outflow_end(basic_storm):
    assert hasattr(basic_storm, "outflow_end")
    ts = pandas.Timestamp("2013-05-19 11:50")
    assert basic_storm.outflow_end.strftime("%X %x") == ts.strftime("%X %x")


def test_Storm_duration_hours(basic_storm):
    assert hasattr(basic_storm, "duration_hours")
    assert abs(basic_storm.duration_hours - 5.75) < 0.0001


def test_Storm_antecedent_period_days(basic_storm):
    assert hasattr(basic_storm, "antecedent_period_days")
    assert abs(basic_storm.antecedent_period_days - 0.17708333) < 0.0001


def test_Storm_peak_precip_intensity(basic_storm):
    assert hasattr(basic_storm, "peak_precip_intensity")
    assert abs(basic_storm.peak_precip_intensity - 1.32) < 0.0001


def test_Storm_peak_inflow(basic_storm):
    assert hasattr(basic_storm, "peak_inflow")
    assert abs(basic_storm.peak_inflow - 123.0) < 0.0001


def test_Storm_peak_outflow(basic_storm):
    assert hasattr(basic_storm, "peak_outflow")
    assert abs(basic_storm.peak_outflow - 41.0) < 0.0001


def test_Storm_peak_precip_intensity_time(basic_storm):
    assert hasattr(basic_storm, "peak_precip_intensity_time")
    ts = pandas.Timestamp("2013-05-19 08:00")
    assert basic_storm.peak_precip_intensity_time.strftime("%X %x") == ts.strftime("%X %x")


def test_Storm_peak_inflow_time(basic_storm):
    assert hasattr(basic_storm, "peak_inflow_time")
    ts = pandas.Timestamp("2013-05-19 08:35")
    assert basic_storm.peak_inflow_time.strftime("%X %x") == ts.strftime("%X %x")


def test_Storm_peak_outflow_time(basic_storm):
    assert hasattr(basic_storm, "peak_outflow_time")
    ts = pandas.Timestamp("2013-05-19 08:45")
    assert basic_storm.peak_outflow_time.strftime("%X %x") == ts.strftime("%X %x")


def test_Storm_centroid_precip(basic_storm):
    assert hasattr(basic_storm, "centroid_precip_time")
    ts = pandas.Timestamp("2013-05-19 07:41:58.705023")
    assert basic_storm.centroid_precip_time.strftime("%X %x") == ts.strftime("%X %x")


def test_Storm_centroid_inflow(basic_storm):
    assert hasattr(basic_storm, "centroid_inflow_time")
    ts = pandas.Timestamp("2013-05-19 08:44:31.252553")
    assert basic_storm.centroid_inflow_time.strftime("%X %x") == ts.strftime("%X %x")


def test_Storm_centroid_outflow(basic_storm):
    assert hasattr(basic_storm, "centroid_outflow_time")
    ts = pandas.Timestamp("2013-05-19 08:59:51.675231")
    assert basic_storm.centroid_outflow_time.strftime("%X %x") == ts.strftime("%X %x")


def test_Storm_total_precip_depth(basic_storm):
    assert hasattr(basic_storm, "total_precip_depth")
    assert abs(basic_storm.total_precip_depth - 1.39) < 0.0001


def test_Storm_total_inflow_volume(basic_storm):
    assert hasattr(basic_storm, "total_inflow_volume")
    assert abs(basic_storm.total_inflow_volume - 876600) < 0.0001


def test_Storm_total_outflow_volume(basic_storm):
    assert hasattr(basic_storm, "total_outflow_volume")
    assert abs(basic_storm.total_outflow_volume - 291900) < 0.0001


def test_Storm_peak_lag_hours(basic_storm):
    assert hasattr(basic_storm, "peak_lag_hours")
    assert abs(basic_storm.peak_lag_hours - 0.1666666667) < 0.0001


def test_Storm_centroid_lag_hours(basic_storm):
    assert hasattr(basic_storm, "centroid_lag_hours")
    assert abs(basic_storm.centroid_lag_hours - 0.255672966) < 0.0001


def test_Storm_summary_dict(basic_storm):
    assert hasattr(basic_storm, "summary_dict")
    assert isinstance(basic_storm.summary_dict, dict)
    known_keys = [
        "Storm Number",
        "Antecedent Days",
        "Start Date",
        "End Date",
        "Duration Hours",
        "Peak Precip Intensity",
        "Total Precip Depth",
        "Total Inflow Volume",
        "Peak Inflow",
        "Total Outflow Volume",
        "Peak Outflow",
        "Peak Lag Hours",
        "Centroid Lag Hours",
        "Season",
    ]
    keys = list(basic_storm.summary_dict.keys())
    assert sorted(keys) == sorted(known_keys)


def test_Storm_is_small(basic_storm):
    assert basic_storm.is_small(minprecip=5.0)
    assert not basic_storm.is_small(minprecip=1.0)


@pytest.fixture
def single_storm():
    pyplot.rcdefaults()
    storm_file = helpers.test_data_path("teststorm_simple.csv")
    orig_record = (
        pandas.read_csv(storm_file, index_col="date", parse_dates=True)
        .resample(FIVEMIN)
        .asfreq()
        .fillna(0)
    )

    hr = hydro.HydroRecord(
        orig_record,
        precipcol="rain",
        inflowcol="influent",
        outflowcol=None,
        outputfreqMinutes=5,
        minprecip=0,
        intereventHours=3,
    )

    return hr.storms[2]


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_plot_storm_summary(single_storm):
    labels = {
        single_storm.inflowcol: "Effluent (l/s)",
        single_storm.precipcol: "Precip Depth (mm)",
    }
    with seaborn.axes_style("ticks"):
        fig, _, labels = single_storm.summaryPlot(outflow=False, serieslabels=labels)
    return fig


@pytest.fixture
def drainage_area():
    return hydro.DrainageArea(100, 75, 10)


def test_da_total_area(drainage_area):
    assert drainage_area.total_area == 100


def test_da_imp_area(drainage_area):
    assert drainage_area.imp_area == 75


def test_da_bmp_area(drainage_area):
    assert drainage_area.bmp_area == 10


@pytest.mark.parametrize(
    ("conversion", "factor", "expected"),
    [(1.0, 1.0, 82.50), (1.0, 0.9, 75.25), (2.0, 1.0, 165.00), (2.0, 0.9, 150.50)],
)
def test_simple_method(drainage_area, conversion, factor, expected):
    depth = 1
    result = drainage_area.simple_method(depth, annual_factor=factor, volume_conversion=conversion)
    area = drainage_area.total_area + drainage_area.bmp_area
    storm_volume = depth * conversion * factor * area
    assert abs(expected - result) < 0.001
    assert storm_volume > result



import warnings
from datetime import datetime

import pandas
import pytest

from wqio.tests import helpers
from wqio.utils import dateutils


@pytest.mark.parametrize(
    "datemaker", [pandas.Timestamp, lambda x: datetime.strptime(x, "%Y-%m-%d")]
)
@pytest.mark.parametrize(
    ("datestring", "expected"),
    [
        ("1998-12-25", "winter"),
        ("2015-03-22", "spring"),
        ("1965-07-04", "summer"),
        ("1982-11-24", "autumn"),
    ],
)
def test_getSeason(datemaker, datestring, expected):
    date = datemaker(datestring)
    season = dateutils.get_season(date)
    assert season == expected


@pytest.mark.parametrize(
    ("row", "expected", "error"),
    [
        ({"sampledate": "2012-05-25", "sampletime": "16:54"}, "2012-05-25 16:54", None),
        ({"sampledate": None, "sampletime": "16:54"}, "1901-01-01 16:54", None),
        ({"sampledate": "2012-05-25", "sampletime": None}, "2012-05-25 00:00", None),
        ({"sampledate": None, "sampletime": None}, "1901-01-01 00:00", None),
        ({"sampledate": "garbage", "sampletime": "16:54"}, "1901-01-01 16:54", None),
        (
            {"sampledate": "2012-05-25", "sampletime": "garbage"},
            "2012-05-25 00:00",
            None,
        ),
        ({"sampledate": "garbage", "sampletime": "garbage"}, "1901-01-01 00:00", None),
    ],
)
def test_makeTimestamp_basic(row, expected, error):
    with warnings.catch_warnings(), helpers.raises(error):
        warnings.simplefilter("always")
        tstamp = dateutils.make_timestamp(row)
        assert tstamp == pandas.Timestamp(expected)


def test_makeTimestamp_customcols():
    row = {"mydate": "2012-05-25", "mytime": "16:54"}
    tstamp = dateutils.make_timestamp(row, datecol="mydate", timecol="mytime")
    assert tstamp == pandas.Timestamp("2012-05-25 16:54")


@pytest.mark.parametrize(
    "datemaker", [pandas.Timestamp, lambda x: datetime.strptime(x, "%Y-%m-%d")]
)
@pytest.mark.parametrize("datestring", ["2005-10-02", "2006-09-02"])
def test_get_wateryear(datemaker, datestring):
    date = datemaker(datestring)
    wateryear = dateutils.get_wateryear(date)
    assert wateryear == "2005/2006"



from datetime import datetime

import numpy
import pandas
import pytest
from matplotlib import pyplot

from wqio import validate
from wqio.tests import helpers


@pytest.mark.parametrize(
    ("fname", "expected", "error_to_raise"),
    [
        ("BMPdata", "bmpdata.zip", None),
        ("bmpdata.csv", "bmpdata.zip", None),
        ("bmpdata.zip", "bmpdata.zip", None),
        ("NSQD", "nsqd.zip", None),
        ("NSQD.csv", "nsqd.zip", None),
        ("NsqD.zip", "nsqd.zip", None),
        ("CvC", "cvc.zip", None),
        ("CVC.csv", "cvc.zip", None),
        ("cVc.zip", "cvc.zip", None),
        ("junk", None, ValueError),
    ],
)
def test_dataset(fname, expected, error_to_raise):
    with helpers.raises(error_to_raise):
        assert expected == validate.dataset(fname)


@pytest.mark.parametrize(
    ("value", "expected", "error_to_raise"),
    [
        (datetime(2015, 1, 5), pandas.Timestamp("2015-01-05"), None),
        ("2015-01-05", pandas.Timestamp("2015-01-05"), None),
        (pandas.Timestamp("2015-01-05"), pandas.Timestamp("2015-01-05"), None),
        ("junk;", None, ValueError),
    ],
)
def test_timestamp(value, expected, error_to_raise):
    with helpers.raises(error_to_raise):
        assert validate.timestamp(value) == expected


def test_axes_invalid():
    with helpers.raises(ValueError):
        validate.axes("junk")


def test_axes_with_ax():
    fig, ax = pyplot.subplots()
    fig1, ax1 = validate.axes(ax)
    assert isinstance(ax1, pyplot.Axes)
    assert isinstance(fig1, pyplot.Figure)
    assert ax1 is ax
    assert fig1 is fig


def test_axes_with_None():
    fig1, ax1 = validate.axes(None)
    assert isinstance(ax1, pyplot.Axes)
    assert isinstance(fig1, pyplot.Figure)


def test_axes_fallback_current():
    fig, ax = pyplot.subplots()
    fig1, ax1 = validate.axes(None, fallback="current")
    assert isinstance(ax1, pyplot.Axes)
    assert isinstance(fig1, pyplot.Figure)
    assert ax1 is ax
    assert fig1 is fig


@pytest.fixture
def multiindex_df():
    dates = range(5)
    params = list("ABCDE")
    locations = ["Inflow", "Outflow"]
    index = pandas.MultiIndex.from_product(
        [dates, params, locations], names=["date", "param", "loc"]
    )
    data = pandas.DataFrame(numpy.random.normal(size=len(index)), index=index)
    return data


@pytest.mark.parametrize(
    ("level", "expected", "error_to_raise"),
    [("param", "A", None), ("date", None, ValueError)],
)
@helpers.seed
def test_getUniqueDataframeIndexVal(multiindex_df, level, expected, error_to_raise):
    data = multiindex_df.xs("A", level="param", drop_level=False)
    with helpers.raises(error_to_raise):
        result = validate.single_value_in_index(data, level)
        assert result == expected


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        (None, []),
        ("", []),
        (1, [1]),
        ("abc", ["abc"]),
        (numpy.array([1, 2, 3]), [1, 2, 3]),
        ([1, 2, 3], [1, 2, 3]),
        (0, [0]),
    ],
)
def test_at_least_empty_list(value, expected):
    result = validate.at_least_empty_list(value)
    assert result == expected


@pytest.mark.parametrize(
    ("value", "options", "expected", "error_to_raise"),
    [
        (None, None, {}, None),
        ("", None, {}, None),
        ({"a": 1, "b": "bee"}, None, {"a": 1, "b": "bee"}, None),
        (None, {"a": 2}, {"a": 2}, None),
        ("", {"a": 2}, {"a": 2}, None),
        ({"a": 1, "b": "bee"}, {"a": 2}, {"a": 2, "b": "bee"}, None),
        ("a", {"b": 1}, None, ValueError),
    ],
)
def test_at_least_empty_dict(value, options, expected, error_to_raise):
    with helpers.raises(error_to_raise):
        if options is None:
            result = validate.at_least_empty_dict(value)
        else:
            result = validate.at_least_empty_dict(value, **options)

        assert result == expected


@pytest.mark.parametrize(
    ("value", "error_to_raise"),
    [
        ("x", None),
        ("y", None),
        ("both", None),
        (None, None),
        ("abc", ValueError),
        (1, ValueError),
    ],
)
def test_fit_arguments(value, error_to_raise):
    with helpers.raises(error_to_raise):
        result = validate.fit_arguments(value, "example arg name")
        assert result == value



import os
from pathlib import Path
from urllib import request
from zipfile import ZipFile

from wqio import validate


def download(dataset, year=None, redownload=True, data_dir=None):
    fname = validate.dataset(dataset)

    tag = "main" if year is None else f"v{year:d}"

    url_template = (
        "https://github.com/Geosyntec/water-quality-datasets/blob/{tag:s}/data/{fname:s}?raw=true"
    )
    src_url = url_template.format(tag=tag, fname=fname)

    if data_dir is None:
        base_dir = Path(os.environ.get("WQ_DATA", "~/.wq-data"))
        data_dir = base_dir.expanduser().absolute() / tag
    else:
        data_dir = Path(data_dir)

    data_dir.mkdir(exist_ok=True, parents=True)
    dst_path = data_dir / fname
    if not dst_path.exists() or redownload:
        request.urlretrieve(src_url, dst_path)

        with ZipFile(dst_path, "r") as zip_ref:
            zip_ref.extractall(data_dir)

    return dst_path.parent / f"{dst_path.stem}.csv"



# Setup script for the wqio package
#
# Usage: python setup.py install
#
from setuptools import find_packages, setup

DESCRIPTION = "wqio: Water Quality Inflow/Outflow"
LONG_DESCRIPTION = DESCRIPTION
NAME = "wqio"
VERSION = "0.6.1"
AUTHOR = "Paul Hobson (Herrera Environmental Consultants)"
AUTHOR_EMAIL = "phobson@herrerainc.com"
URL = "https://github.com/International-BMP-Database/wqio"
DOWNLOAD_URL = "https://github.com/International-BMP-Database/wqio/archive/main.zip"
LICENSE = "BSD 3-clause"
PACKAGES = find_packages()
PLATFORMS = "Python 3.9 and later."
CLASSIFIERS = [
    "License :: OSI Approved :: BSD License",
    "Operating System :: OS Independent",
    "Programming Language :: Python",
    "Intended Audience :: Science/Research",
    "Topic :: Software Development :: Libraries :: Python Modules",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3.10",
    "Programming Language :: Python :: 3.11",
    "Programming Language :: Python :: 3.12",
]
INSTALL_REQUIRES = [
    "numpy",
    "scipy",
    "matplotlib",
    "pandas",
    "statsmodels",
    "seaborn",
    "probscale",
]
PACKAGE_DATA = {
    "wqio.tests._data": ["*.csv", "*.tex", "matplotlibrc"],
    "wqio.tests._baseline_images.features_tests": ["*png"],
    "wqio.tests._baseline_images.hydro_tests": ["*png"],
    "wqio.tests._baseline_images.samples_tests": ["*png"],
    "wqio.tests._baseline_images.utils_tests.figutils_tests": ["*png"],
}

setup(
    name=NAME,
    version=VERSION,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    url=URL,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    download_url=DOWNLOAD_URL,
    license=LICENSE,
    packages=PACKAGES,
    package_data=PACKAGE_DATA,
    platforms=PLATFORMS,
    classifiers=CLASSIFIERS,
    install_requires=INSTALL_REQUIRES,
)



from io import StringIO
from textwrap import dedent

import numpy
import numpy.testing as nptest
import pandas
import pandas.testing as pdtest
import pytest

from wqio import ros
from wqio.tests import helpers


@pytest.fixture
def basic_data():
    df = (
        helpers.getTestROSData()
        .assign(conc=lambda df: df["res"])
        .assign(censored=lambda df: df["qual"] == "ND")
    )
    return df


@pytest.fixture
def intermediate_data():
    df = pandas.DataFrame(
        [
            {"censored": True, "conc": 5.0, "det_limit_index": 1, "rank": 1},
            {"censored": True, "conc": 5.0, "det_limit_index": 1, "rank": 2},
            {"censored": True, "conc": 5.5, "det_limit_index": 2, "rank": 1},
            {"censored": True, "conc": 5.75, "det_limit_index": 3, "rank": 1},
            {"censored": True, "conc": 9.5, "det_limit_index": 4, "rank": 1},
            {"censored": True, "conc": 9.5, "det_limit_index": 4, "rank": 2},
            {"censored": True, "conc": 11.0, "det_limit_index": 5, "rank": 1},
            {"censored": False, "conc": 2.0, "det_limit_index": 0, "rank": 1},
            {"censored": False, "conc": 4.2, "det_limit_index": 0, "rank": 2},
            {"censored": False, "conc": 4.62, "det_limit_index": 0, "rank": 3},
            {"censored": False, "conc": 5.57, "det_limit_index": 2, "rank": 1},
            {"censored": False, "conc": 5.66, "det_limit_index": 2, "rank": 2},
            {"censored": False, "conc": 5.86, "det_limit_index": 3, "rank": 1},
            {"censored": False, "conc": 6.65, "det_limit_index": 3, "rank": 2},
            {"censored": False, "conc": 6.78, "det_limit_index": 3, "rank": 3},
            {"censored": False, "conc": 6.79, "det_limit_index": 3, "rank": 4},
            {"censored": False, "conc": 7.5, "det_limit_index": 3, "rank": 5},
            {"censored": False, "conc": 7.5, "det_limit_index": 3, "rank": 6},
            {"censored": False, "conc": 7.5, "det_limit_index": 3, "rank": 7},
            {"censored": False, "conc": 8.63, "det_limit_index": 3, "rank": 8},
            {"censored": False, "conc": 8.71, "det_limit_index": 3, "rank": 9},
            {"censored": False, "conc": 8.99, "det_limit_index": 3, "rank": 10},
            {"censored": False, "conc": 9.85, "det_limit_index": 4, "rank": 1},
            {"censored": False, "conc": 10.82, "det_limit_index": 4, "rank": 2},
            {"censored": False, "conc": 11.25, "det_limit_index": 5, "rank": 1},
            {"censored": False, "conc": 11.25, "det_limit_index": 5, "rank": 2},
            {"censored": False, "conc": 12.2, "det_limit_index": 5, "rank": 3},
            {"censored": False, "conc": 14.92, "det_limit_index": 5, "rank": 4},
            {"censored": False, "conc": 16.77, "det_limit_index": 5, "rank": 5},
            {"censored": False, "conc": 17.81, "det_limit_index": 5, "rank": 6},
            {"censored": False, "conc": 19.16, "det_limit_index": 5, "rank": 7},
            {"censored": False, "conc": 19.19, "det_limit_index": 5, "rank": 8},
            {"censored": False, "conc": 19.64, "det_limit_index": 5, "rank": 9},
            {"censored": False, "conc": 20.18, "det_limit_index": 5, "rank": 10},
            {"censored": False, "conc": 22.97, "det_limit_index": 5, "rank": 11},
        ]
    )

    return df


@pytest.fixture
def advanced_data():
    df = pandas.DataFrame(
        [
            {
                "Zprelim": -1.44562021,
                "censored": True,
                "conc": 5.0,
                "det_limit_index": 1,
                "plot_pos": 0.07414187,
                "rank": 1,
            },
            {
                "Zprelim": -1.22010353,
                "censored": True,
                "conc": 5.0,
                "det_limit_index": 1,
                "plot_pos": 0.11121281,
                "rank": 2,
            },
            {
                "Zprelim": -1.04382253,
                "censored": True,
                "conc": 5.5,
                "det_limit_index": 2,
                "plot_pos": 0.14828375,
                "rank": 1,
            },
            {
                "Zprelim": -1.04382253,
                "censored": True,
                "conc": 5.75,
                "det_limit_index": 3,
                "plot_pos": 0.14828375,
                "rank": 1,
            },
            {
                "Zprelim": -0.81095536,
                "censored": True,
                "conc": 9.5,
                "det_limit_index": 4,
                "plot_pos": 0.20869565,
                "rank": 1,
            },
            {
                "Zprelim": -0.40467790,
                "censored": True,
                "conc": 9.5,
                "det_limit_index": 4,
                "plot_pos": 0.34285714,
                "rank": 2,
            },
            {
                "Zprelim": -0.20857169,
                "censored": True,
                "conc": 11.0,
                "det_limit_index": 5,
                "plot_pos": 0.41739130,
                "rank": 1,
            },
            {
                "Zprelim": -1.59276546,
                "censored": False,
                "conc": 2.0,
                "det_limit_index": 0,
                "plot_pos": 0.05560640,
                "rank": 1,
            },
            {
                "Zprelim": -1.22010353,
                "censored": False,
                "conc": 4.2,
                "det_limit_index": 0,
                "plot_pos": 0.11121281,
                "rank": 2,
            },
            {
                "Zprelim": -0.96681116,
                "censored": False,
                "conc": 4.62,
                "det_limit_index": 0,
                "plot_pos": 0.16681922,
                "rank": 3,
            },
            {
                "Zprelim": -0.68351863,
                "censored": False,
                "conc": 5.57,
                "det_limit_index": 2,
                "plot_pos": 0.24713958,
                "rank": 1,
            },
            {
                "Zprelim": -0.60721672,
                "censored": False,
                "conc": 5.66,
                "det_limit_index": 2,
                "plot_pos": 0.27185354,
                "rank": 2,
            },
            {
                "Zprelim": -0.44953240,
                "censored": False,
                "conc": 5.86,
                "det_limit_index": 3,
                "plot_pos": 0.32652381,
                "rank": 1,
            },
            {
                "Zprelim": -0.36788328,
                "censored": False,
                "conc": 6.65,
                "det_limit_index": 3,
                "plot_pos": 0.35648013,
                "rank": 2,
            },
            {
                "Zprelim": -0.28861907,
                "censored": False,
                "conc": 6.78,
                "det_limit_index": 3,
                "plot_pos": 0.38643644,
                "rank": 3,
            },
            {
                "Zprelim": -0.21113039,
                "censored": False,
                "conc": 6.79,
                "det_limit_index": 3,
                "plot_pos": 0.41639276,
                "rank": 4,
            },
            {
                "Zprelim": -0.13489088,
                "censored": False,
                "conc": 7.5,
                "det_limit_index": 3,
                "plot_pos": 0.44634907,
                "rank": 5,
            },
            {
                "Zprelim": -0.05942854,
                "censored": False,
                "conc": 7.5,
                "det_limit_index": 3,
                "plot_pos": 0.47630538,
                "rank": 6,
            },
            {
                "Zprelim": 0.015696403,
                "censored": False,
                "conc": 7.5,
                "det_limit_index": 3,
                "plot_pos": 0.50626170,
                "rank": 7,
            },
            {
                "Zprelim": 0.090910169,
                "censored": False,
                "conc": 8.63,
                "det_limit_index": 3,
                "plot_pos": 0.53621801,
                "rank": 8,
            },
            {
                "Zprelim": 0.166642511,
                "censored": False,
                "conc": 8.71,
                "det_limit_index": 3,
                "plot_pos": 0.56617432,
                "rank": 9,
            },
            {
                "Zprelim": 0.243344267,
                "censored": False,
                "conc": 8.99,
                "det_limit_index": 3,
                "plot_pos": 0.59613064,
                "rank": 10,
            },
            {
                "Zprelim": 0.374443298,
                "censored": False,
                "conc": 9.85,
                "det_limit_index": 4,
                "plot_pos": 0.64596273,
                "rank": 1,
            },
            {
                "Zprelim": 0.428450751,
                "censored": False,
                "conc": 10.82,
                "det_limit_index": 4,
                "plot_pos": 0.66583850,
                "rank": 2,
            },
            {
                "Zprelim": 0.558957865,
                "censored": False,
                "conc": 11.25,
                "det_limit_index": 5,
                "plot_pos": 0.71190476,
                "rank": 1,
            },
            {
                "Zprelim": 0.637484160,
                "censored": False,
                "conc": 11.25,
                "det_limit_index": 5,
                "plot_pos": 0.73809523,
                "rank": 2,
            },
            {
                "Zprelim": 0.720156617,
                "censored": False,
                "conc": 12.2,
                "det_limit_index": 5,
                "plot_pos": 0.76428571,
                "rank": 3,
            },
            {
                "Zprelim": 0.808074633,
                "censored": False,
                "conc": 14.92,
                "det_limit_index": 5,
                "plot_pos": 0.79047619,
                "rank": 4,
            },
            {
                "Zprelim": 0.902734791,
                "censored": False,
                "conc": 16.77,
                "det_limit_index": 5,
                "plot_pos": 0.81666666,
                "rank": 5,
            },
            {
                "Zprelim": 1.006269985,
                "censored": False,
                "conc": 17.81,
                "det_limit_index": 5,
                "plot_pos": 0.84285714,
                "rank": 6,
            },
            {
                "Zprelim": 1.121900467,
                "censored": False,
                "conc": 19.16,
                "det_limit_index": 5,
                "plot_pos": 0.86904761,
                "rank": 7,
            },
            {
                "Zprelim": 1.254875912,
                "censored": False,
                "conc": 19.19,
                "det_limit_index": 5,
                "plot_pos": 0.89523809,
                "rank": 8,
            },
            {
                "Zprelim": 1.414746425,
                "censored": False,
                "conc": 19.64,
                "det_limit_index": 5,
                "plot_pos": 0.92142857,
                "rank": 9,
            },
            {
                "Zprelim": 1.622193585,
                "censored": False,
                "conc": 20.18,
                "det_limit_index": 5,
                "plot_pos": 0.94761904,
                "rank": 10,
            },
            {
                "Zprelim": 1.939989611,
                "censored": False,
                "conc": 22.97,
                "det_limit_index": 5,
                "plot_pos": 0.97380952,
                "rank": 11,
            },
        ]
    )

    return df


@pytest.fixture
def basic_cohn():
    cohn = pandas.DataFrame(
        [
            {
                "lower_dl": 2.0,
                "ncen_equal": 0.0,
                "nobs_below": 0.0,
                "nuncen_above": 3.0,
                "prob_exceedance": 1.0,
                "upper_dl": 5.0,
            },
            {
                "lower_dl": 5.0,
                "ncen_equal": 2.0,
                "nobs_below": 5.0,
                "nuncen_above": 0.0,
                "prob_exceedance": 0.77757437070938218,
                "upper_dl": 5.5,
            },
            {
                "lower_dl": 5.5,
                "ncen_equal": 1.0,
                "nobs_below": 6.0,
                "nuncen_above": 2.0,
                "prob_exceedance": 0.77757437070938218,
                "upper_dl": 5.75,
            },
            {
                "lower_dl": 5.75,
                "ncen_equal": 1.0,
                "nobs_below": 9.0,
                "nuncen_above": 10.0,
                "prob_exceedance": 0.7034324942791762,
                "upper_dl": 9.5,
            },
            {
                "lower_dl": 9.5,
                "ncen_equal": 2.0,
                "nobs_below": 21.0,
                "nuncen_above": 2.0,
                "prob_exceedance": 0.37391304347826088,
                "upper_dl": 11.0,
            },
            {
                "lower_dl": 11.0,
                "ncen_equal": 1.0,
                "nobs_below": 24.0,
                "nuncen_above": 11.0,
                "prob_exceedance": 0.31428571428571428,
                "upper_dl": numpy.inf,
            },
            {
                "lower_dl": numpy.nan,
                "ncen_equal": numpy.nan,
                "nobs_below": numpy.nan,
                "nuncen_above": numpy.nan,
                "prob_exceedance": 0.0,
                "upper_dl": numpy.nan,
            },
        ]
    )
    return cohn


@pytest.fixture
def expected_sorted():
    expected_sorted = pandas.DataFrame(
        [
            {"censored": True, "conc": 5.0},
            {"censored": True, "conc": 5.0},
            {"censored": True, "conc": 5.5},
            {"censored": True, "conc": 5.75},
            {"censored": True, "conc": 9.5},
            {"censored": True, "conc": 9.5},
            {"censored": True, "conc": 11.0},
            {"censored": False, "conc": 2.0},
            {"censored": False, "conc": 4.2},
            {"censored": False, "conc": 4.62},
            {"censored": False, "conc": 5.57},
            {"censored": False, "conc": 5.66},
            {"censored": False, "conc": 5.86},
            {"censored": False, "conc": 6.65},
            {"censored": False, "conc": 6.78},
            {"censored": False, "conc": 6.79},
            {"censored": False, "conc": 7.5},
            {"censored": False, "conc": 7.5},
            {"censored": False, "conc": 7.5},
            {"censored": False, "conc": 8.63},
            {"censored": False, "conc": 8.71},
            {"censored": False, "conc": 8.99},
            {"censored": False, "conc": 9.85},
            {"censored": False, "conc": 10.82},
            {"censored": False, "conc": 11.25},
            {"censored": False, "conc": 11.25},
            {"censored": False, "conc": 12.2},
            {"censored": False, "conc": 14.92},
            {"censored": False, "conc": 16.77},
            {"censored": False, "conc": 17.81},
            {"censored": False, "conc": 19.16},
            {"censored": False, "conc": 19.19},
            {"censored": False, "conc": 19.64},
            {"censored": False, "conc": 20.18},
            {"censored": False, "conc": 22.97},
        ]
    )[["conc", "censored"]]
    return expected_sorted


@pytest.fixture
def expected_cohn():
    final_cols = [
        "lower_dl",
        "upper_dl",
        "nuncen_above",
        "nobs_below",
        "ncen_equal",
        "prob_exceedance",
    ]
    cohn = pandas.DataFrame(
        [
            {
                "lower_dl": 2.0,
                "ncen_equal": 0.0,
                "nobs_below": 0.0,
                "nuncen_above": 3.0,
                "prob_exceedance": 1.0,
                "upper_dl": 5.0,
            },
            {
                "lower_dl": 5.0,
                "ncen_equal": 2.0,
                "nobs_below": 5.0,
                "nuncen_above": 0.0,
                "prob_exceedance": 0.77757437070938218,
                "upper_dl": 5.5,
            },
            {
                "lower_dl": 5.5,
                "ncen_equal": 1.0,
                "nobs_below": 6.0,
                "nuncen_above": 2.0,
                "prob_exceedance": 0.77757437070938218,
                "upper_dl": 5.75,
            },
            {
                "lower_dl": 5.75,
                "ncen_equal": 1.0,
                "nobs_below": 9.0,
                "nuncen_above": 10.0,
                "prob_exceedance": 0.7034324942791762,
                "upper_dl": 9.5,
            },
            {
                "lower_dl": 9.5,
                "ncen_equal": 2.0,
                "nobs_below": 21.0,
                "nuncen_above": 2.0,
                "prob_exceedance": 0.37391304347826088,
                "upper_dl": 11.0,
            },
            {
                "lower_dl": 11.0,
                "ncen_equal": 1.0,
                "nobs_below": 24.0,
                "nuncen_above": 11.0,
                "prob_exceedance": 0.31428571428571428,
                "upper_dl": numpy.inf,
            },
            {
                "lower_dl": numpy.nan,
                "ncen_equal": numpy.nan,
                "nobs_below": numpy.nan,
                "nuncen_above": numpy.nan,
                "prob_exceedance": 0.0,
                "upper_dl": numpy.nan,
            },
        ]
    )[final_cols]
    return cohn


def test__ros_sort_baseline(basic_data, expected_sorted):
    result = ros._ros_sort(basic_data, "conc", "censored")
    pdtest.assert_frame_equal(result, expected_sorted, rtol=1e-5)


def test__ros_sort_warning(basic_data, expected_sorted):
    df = basic_data.copy()
    max_row = df["conc"].idxmax()
    df.loc[max_row, "censored"] = True
    with pytest.warns(UserWarning):
        result = ros._ros_sort(df, "conc", "censored", warn=True)
        pdtest.assert_frame_equal(result, expected_sorted.iloc[:-1], rtol=1e-5)


def test_cohn_numbers_baseline(basic_data, expected_cohn):
    result = ros.cohn_numbers(basic_data, result="conc", censorship="censored")
    pdtest.assert_frame_equal(result, expected_cohn, rtol=1e-5)


def test_cohn_numbers_no_NDs(basic_data):
    result = ros.cohn_numbers(basic_data.assign(qual=False), result="conc", censorship="qual")
    assert result.shape == (0, 6)


def test__detection_limit_index_empty():
    empty_cohn = pandas.DataFrame(numpy.empty((0, 7)))
    assert ros._detection_limit_index(None, empty_cohn) == 0


@pytest.mark.parametrize(("value", "expected"), [(3.5, 0), (6.0, 3), (12.0, 5)])
def test__detection_limit_index_populated(value, expected, basic_cohn):
    result = ros._detection_limit_index(value, basic_cohn)
    assert result == expected


def test__detection_limit_index_out_of_bounds(basic_cohn):
    with helpers.raises(IndexError):
        ros._detection_limit_index(0, basic_cohn)


def test__ros_group_rank():
    df = pandas.DataFrame(
        {
            "dl_idx": [1] * 12,
            "params": list("AABCCCDE") + list("DCBA"),
            "values": list(range(12)),
        }
    )

    result = ros._ros_group_rank(df, "dl_idx", "params")
    expected = pandas.Series([1, 2, 1, 1, 2, 3, 1, 1, 2, 4, 2, 3], name="rank")
    pdtest.assert_series_equal(result, expected)


@pytest.mark.parametrize(
    ("row", "expected"),
    [
        ({"censored": False, "det_limit_index": 2, "rank": 1}, 0.24713958810068648),
        ({"censored": False, "det_limit_index": 2, "rank": 12}, 0.51899313501144173),
        ({"censored": True, "det_limit_index": 5, "rank": 4}, 1.3714285714285714),
        ({"censored": True, "det_limit_index": 4, "rank": 2}, 0.41739130434782606),
    ],
)
def test__ros_plot_pos(row, expected, basic_cohn):
    result = ros._ros_plot_pos(row, "censored", basic_cohn)
    assert (abs(result - expected) / expected) < 0.00001


def test__norm_plot_pos():
    result = ros._norm_plot_pos([1, 2, 3, 4])
    expected = numpy.array([0.159104, 0.385452, 0.614548, 0.840896])
    nptest.assert_array_almost_equal(result, expected)


def test_plotting_positions(intermediate_data, basic_cohn):
    results = ros.plotting_positions(intermediate_data, "censored", basic_cohn)
    expected = numpy.array(
        [
            0.07414188,
            0.11121281,
            0.14828375,
            0.14828375,
            0.20869565,
            0.34285714,
            0.41739130,
            0.05560641,
            0.11121281,
            0.16681922,
            0.24713959,
            0.27185355,
            0.32652382,
            0.35648013,
            0.38643645,
            0.41639276,
            0.44634907,
            0.47630539,
            0.50626170,
            0.53621802,
            0.56617433,
            0.59613064,
            0.64596273,
            0.66583851,
            0.71190476,
            0.73809524,
            0.76428571,
            0.79047619,
            0.81666667,
            0.84285714,
            0.86904762,
            0.89523810,
            0.92142857,
            0.94761905,
            0.97380952,
        ]
    )
    nptest.assert_array_almost_equal(results, expected)


def test__ros_estimate(advanced_data):
    expected = numpy.array(
        [
            3.11279729,
            3.60634338,
            4.04602788,
            4.04602788,
            4.71008116,
            6.14010906,
            6.97841457,
            2.00000000,
            4.20000000,
            4.62000000,
            5.57000000,
            5.66000000,
            5.86000000,
            6.65000000,
            6.78000000,
            6.79000000,
            7.50000000,
            7.50000000,
            7.50000000,
            8.63000000,
            8.71000000,
            8.99000000,
            9.85000000,
            10.82000000,
            11.25000000,
            11.25000000,
            12.20000000,
            14.92000000,
            16.77000000,
            17.81000000,
            19.16000000,
            19.19000000,
            19.64000000,
            20.18000000,
            22.97,
        ]
    )
    df = advanced_data.pipe(ros._ros_estimate, "conc", "censored", numpy.log, numpy.exp)
    result = df["final"].values
    nptest.assert_array_almost_equal(result, expected)


def test__do_ros_basic(basic_data):
    expected = numpy.array(
        [
            3.11279729,
            3.60634338,
            4.04602788,
            4.04602788,
            4.71008116,
            6.14010906,
            6.97841457,
            2.00000000,
            4.20000000,
            4.62000000,
            5.57000000,
            5.66000000,
            5.86000000,
            6.65000000,
            6.78000000,
            6.79000000,
            7.50000000,
            7.50000000,
            7.50000000,
            8.63000000,
            8.71000000,
            8.99000000,
            9.85000000,
            10.82000000,
            11.25000000,
            11.25000000,
            12.20000000,
            14.92000000,
            16.77000000,
            17.81000000,
            19.16000000,
            19.19000000,
            19.64000000,
            20.18000000,
            22.97,
        ]
    )

    df = basic_data.pipe(ros._do_ros, "conc", "censored", numpy.log, numpy.exp)
    result = df["final"].values
    nptest.assert_array_almost_equal(result, expected)


def test__do_ros_basic_with_floor(basic_data):
    expected = numpy.array(
        [
            5.00000000,
            5.00000000,
            5.00000000,
            5.00000000,
            5.00000000,
            6.14010906,
            6.97841457,
            5.00000000,
            5.00000000,
            5.00000000,
            5.57000000,
            5.66000000,
            5.86000000,
            6.65000000,
            6.78000000,
            6.79000000,
            7.50000000,
            7.50000000,
            7.50000000,
            8.63000000,
            8.71000000,
            8.99000000,
            9.85000000,
            10.82000000,
            11.25000000,
            11.25000000,
            12.20000000,
            14.92000000,
            16.77000000,
            17.81000000,
            19.16000000,
            19.19000000,
            19.64000000,
            20.18000000,
            22.97,
        ]
    )

    df = basic_data.pipe(ros._do_ros, "conc", "censored", numpy.log, numpy.exp, floor=5)
    result = df["final"].values
    nptest.assert_array_almost_equal(result, expected)


def test__do_ros_all_equal_some_cen():
    expected = [0.4, 0.4, 0.4, 0.4, 0.4, 0.4, 0.4, 0.4]
    result = (
        pandas.DataFrame(
            {
                "res": numpy.array(expected),
                "cen": numpy.array([True, True, True, True, True, False, True, False]),
            }
        )
        .pipe(ros._do_ros, "res", "cen", numpy.log, numpy.exp)["final"]
        .values
    )
    nptest.assert_array_almost_equal(result, expected)


@pytest.mark.parametrize(
    ["as_obj", "expected"],
    [(True, {"enough_uncensored": True, "not_too_many_censored": True}), (False, True)],
)
def test_is_valid_to_ros(basic_data, as_obj, expected):
    result = ros.is_valid_to_ros(basic_data, "censored", as_obj=as_obj)
    assert result == expected


class HelselAppendixB:
    """
    Appendix B dataset from "Estimation of Descriptive Statists for
    Multiply Censored Water Quality Data", Water Resources Research,
    Vol 24, No 12, pp 1997 - 2004. December 1988.
    """

    decimal = 2
    res = numpy.array(
        [
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            10.0,
            10.0,
            10.0,
            3.0,
            7.0,
            9.0,
            12.0,
            15.0,
            20.0,
            27.0,
            33.0,
            50.0,
        ]
    )
    cen = numpy.array(
        [
            True,
            True,
            True,
            True,
            True,
            True,
            True,
            True,
            True,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
        ]
    )
    rescol = "obs"
    cencol = "cen"
    df = pandas.DataFrame({rescol: res, cencol: cen})
    values = numpy.array(
        [
            0.47,
            0.85,
            1.11,
            1.27,
            1.76,
            2.34,
            2.50,
            3.00,
            3.03,
            4.80,
            7.00,
            9.00,
            12.0,
            15.0,
            20.0,
            27.0,
            33.0,
            50.0,
        ]
    )

    cohn = pandas.DataFrame(
        {
            "nuncen_above": numpy.array([3.0, 6.0, numpy.nan]),
            "nobs_below": numpy.array([6.0, 12.0, numpy.nan]),
            "ncen_equal": numpy.array([6.0, 3.0, numpy.nan]),
            "prob_exceedance": numpy.array([0.5555, 0.3333, 0.0]),
        }
    )


class HelselArsenic:
    """
    Oahu arsenic data from Nondetects and Data Analysis by
    Dennis R. Helsel (John Wiley, 2005)

    Plotting positions are fudged since relative to source data since
    modeled data is what matters and (source data plot positions are
    not uniformly spaced, which seems weird)
    """

    decimal = 2
    res = numpy.array(
        [
            3.2,
            2.8,
            2.0,
            2.0,
            2.0,
            2.0,
            2.0,
            2.0,
            2.0,
            2.0,
            1.7,
            1.5,
            1.0,
            1.0,
            1.0,
            1.0,
            0.9,
            0.9,
            0.7,
            0.7,
            0.6,
            0.5,
            0.5,
            0.5,
        ]
    )

    cen = numpy.array(
        [
            False,
            False,
            True,
            True,
            True,
            True,
            True,
            True,
            True,
            True,
            False,
            False,
            True,
            True,
            True,
            True,
            False,
            True,
            False,
            False,
            False,
            False,
            False,
            False,
        ]
    )
    rescol = "obs"
    cencol = "cen"
    df = pandas.DataFrame({rescol: res, cencol: cen})
    values = numpy.array(
        [
            3.20,
            2.80,
            1.42,
            1.14,
            0.95,
            0.81,
            0.68,
            0.57,
            0.46,
            0.35,
            1.70,
            1.50,
            0.98,
            0.76,
            0.58,
            0.41,
            0.90,
            0.61,
            0.70,
            0.70,
            0.60,
            0.50,
            0.50,
            0.50,
        ]
    )

    cohn = pandas.DataFrame(
        {
            "nuncen_above": numpy.array([6.0, 1.0, 2.0, 2.0, numpy.nan]),
            "nobs_below": numpy.array([0.0, 7.0, 12.0, 22.0, numpy.nan]),
            "ncen_equal": numpy.array([0.0, 1.0, 4.0, 8.0, numpy.nan]),
            "prob_exceedance": numpy.array([1.0, 0.3125, 0.2143, 0.0833, 0.0]),
        }
    )


class RNADAdata:
    decimal = 3
    datastring = StringIO(
        dedent(
            """\
        res cen
        0.090  True
        0.090  True
        0.090  True
        0.101 False
        0.136 False
        0.340 False
        0.457 False
        0.514 False
        0.629 False
        0.638 False
        0.774 False
        0.788 False
        0.900  True
        0.900  True
        0.900  True
        1.000  True
        1.000  True
        1.000  True
        1.000  True
        1.000  True
        1.000 False
        1.000  True
        1.000  True
        1.000  True
        1.000  True
        1.000  True
        1.000  True
        1.000  True
        1.000  True
        1.000  True
        1.000  True
        1.000  True
        1.000  True
        1.100 False
        2.000 False
        2.000 False
        2.404 False
        2.860 False
        3.000 False
        3.000 False
        3.705 False
        4.000 False
        5.000 False
        5.960 False
        6.000 False
        7.214 False
       16.000 False
       17.716 False
       25.000 False
       51.000 False
    """
        )
    )
    rescol = "res"
    cencol = "cen"
    df = pandas.read_csv(datastring, sep=r"\s+")
    values = numpy.array(
        [
            0.01907990,
            0.03826254,
            0.06080717,
            0.10100000,
            0.13600000,
            0.34000000,
            0.45700000,
            0.51400000,
            0.62900000,
            0.63800000,
            0.77400000,
            0.78800000,
            0.08745914,
            0.25257575,
            0.58544205,
            0.01711153,
            0.03373885,
            0.05287083,
            0.07506079,
            0.10081573,
            1.00000000,
            0.13070334,
            0.16539309,
            0.20569039,
            0.25257575,
            0.30725491,
            0.37122555,
            0.44636843,
            0.53507405,
            0.64042242,
            0.76644378,
            0.91850581,
            1.10390531,
            1.10000000,
            2.00000000,
            2.00000000,
            2.40400000,
            2.86000000,
            3.00000000,
            3.00000000,
            3.70500000,
            4.00000000,
            5.00000000,
            5.96000000,
            6.00000000,
            7.21400000,
            16.00000000,
            17.71600000,
            25.00000000,
            51.00000000,
        ]
    )

    cohn = pandas.DataFrame(
        {
            "nuncen_above": numpy.array([9.0, 0.0, 18.0, numpy.nan]),
            "nobs_below": numpy.array([3.0, 15.0, 32.0, numpy.nan]),
            "ncen_equal": numpy.array([3.0, 3.0, 17.0, numpy.nan]),
            "prob_exceedance": numpy.array([0.84, 0.36, 0.36, 0]),
        }
    )


class NoOp_ZeroND:
    decimal = 2
    numpy.random.seed(0)
    N = 20
    res = numpy.random.lognormal(size=N)
    cen = [False] * N
    rescol = "obs"
    cencol = "cen"
    df = pandas.DataFrame({rescol: res, cencol: cen})
    values = numpy.array(
        [
            0.38,
            0.43,
            0.81,
            0.86,
            0.90,
            1.13,
            1.15,
            1.37,
            1.40,
            1.49,
            1.51,
            1.56,
            2.14,
            2.59,
            2.66,
            4.28,
            4.46,
            5.84,
            6.47,
            9.4,
        ]
    )

    cohn = pandas.DataFrame(
        {
            "nuncen_above": numpy.array([]),
            "nobs_below": numpy.array([]),
            "ncen_equal": numpy.array([]),
            "prob_exceedance": numpy.array([]),
        }
    )


class OneND:
    decimal = 3
    res = numpy.array(
        [
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            10.0,
            10.0,
            10.0,
            3.0,
            7.0,
            9.0,
            12.0,
            15.0,
            20.0,
            27.0,
            33.0,
            50.0,
        ]
    )
    cen = numpy.array(
        [
            True,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
        ]
    )
    rescol = "conc"
    cencol = "cen"
    df = pandas.DataFrame({rescol: res, cencol: cen})
    values = numpy.array(
        [
            0.24,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            10.0,
            10.0,
            10.0,
            3.00,
            7.0,
            9.0,
            12.0,
            15.0,
            20.0,
            27.0,
            33.0,
            50.0,
        ]
    )

    cohn = pandas.DataFrame(
        {
            "nuncen_above": numpy.array([17.0, numpy.nan]),
            "nobs_below": numpy.array([1.0, numpy.nan]),
            "ncen_equal": numpy.array([1.0, numpy.nan]),
            "prob_exceedance": numpy.array([0.9444, 0.0]),
        }
    )


class HalfDLs_80pctNDs:
    decimal = 3
    res = numpy.array(
        [
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            10.0,
            10.0,
            10.0,
            3.0,
            7.0,
            9.0,
            12.0,
            15.0,
            20.0,
            27.0,
            33.0,
            50.0,
        ]
    )
    cen = numpy.array(
        [
            True,
            True,
            True,
            True,
            True,
            True,
            True,
            True,
            True,
            True,
            True,
            True,
            True,
            True,
            True,
            False,
            False,
            False,
        ]
    )
    rescol = "value"
    cencol = "qual"
    df = pandas.DataFrame({rescol: res, cencol: cen})
    values = numpy.array(
        [
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            5.0,
            5.0,
            5.0,
            1.5,
            3.5,
            4.5,
            6.0,
            7.5,
            10.0,
            27.0,
            33.0,
            50.0,
        ]
    )

    cohn = pandas.DataFrame(
        {
            "nuncen_above": numpy.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 3.0, numpy.nan]),
            "nobs_below": numpy.array([6.0, 7.0, 8.0, 9.0, 12.0, 13.0, 14.0, 15.0, numpy.nan]),
            "ncen_equal": numpy.array([6.0, 1.0, 1.0, 1.0, 3.0, 1.0, 1.0, 1.0, numpy.nan]),
            "prob_exceedance": numpy.array([0.1667] * 8 + [0.0]),
        }
    )


class HaflDLs_OneUncensored:
    decimal = 3
    res = numpy.array([1.0, 1.0, 12.0, 15.0])
    cen = numpy.array([True, True, True, False])
    rescol = "value"
    cencol = "qual"
    df = pandas.DataFrame({rescol: res, cencol: cen})
    values = numpy.array([0.5, 0.5, 6.0, 15.0])

    cohn = pandas.DataFrame(
        {
            "nuncen_above": numpy.array([0.0, 1.0, numpy.nan]),
            "nobs_below": numpy.array([2.0, 3.0, numpy.nan]),
            "ncen_equal": numpy.array([2.0, 1.0, numpy.nan]),
            "prob_exceedance": numpy.array([0.25, 0.25, 0.0]),
        }
    )


class MaxCen_GT_MaxUncen(HelselAppendixB):
    res = numpy.array(
        [
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            10.0,
            10.0,
            10.0,
            3.0,
            7.0,
            9.0,
            12.0,
            15.0,
            20.0,
            27.0,
            33.0,
            50.0,
            60,
            70,
        ]
    )
    cen = numpy.array(
        [
            True,
            True,
            True,
            True,
            True,
            True,
            True,
            True,
            True,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            True,
            True,
        ]
    )


class OnlyDL_GT_MaxUncen(NoOp_ZeroND):
    numpy.random.seed(0)
    N = 20
    res = [
        0.38,
        0.43,
        0.81,
        0.86,
        0.90,
        1.13,
        1.15,
        1.37,
        1.40,
        1.49,
        1.51,
        1.56,
        2.14,
        2.59,
        2.66,
        4.28,
        4.46,
        5.84,
        6.47,
        9.40,
        10.0,
        10.0,
    ]
    cen = ([False] * N) + [True, True]


@pytest.mark.parametrize(
    "case",
    [
        HelselAppendixB,
        HelselArsenic,
        RNADAdata,
        NoOp_ZeroND,
        OneND,
        HalfDLs_80pctNDs,
        HaflDLs_OneUncensored,
        MaxCen_GT_MaxUncen,
        OnlyDL_GT_MaxUncen,
    ],
)
@pytest.mark.parametrize("as_arrays", [True, False])
def test_ros_from_literature(as_arrays, case):
    if as_arrays:
        result = ros.ROS(case.df[case.rescol], case.df[case.cencol], df=None)
    else:
        result = ros.ROS(case.rescol, case.cencol, df=case.df)

    nptest.assert_array_almost_equal(sorted(result), sorted(case.values), decimal=case.decimal)


@pytest.mark.parametrize(
    "case",
    [
        HelselAppendixB,
        HelselArsenic,
        RNADAdata,
        NoOp_ZeroND,
        OneND,
        HalfDLs_80pctNDs,
        HaflDLs_OneUncensored,
        MaxCen_GT_MaxUncen,
        OnlyDL_GT_MaxUncen,
    ],
)
def test_cohn_from_literature(case):
    cols = ["nuncen_above", "nobs_below", "ncen_equal", "prob_exceedance"]
    result = ros.cohn_numbers(case.df, case.rescol, case.cencol)
    pdtest.assert_frame_equal(result[cols].round(5), case.cohn[cols].round(5), atol=1e-4)



import numpy
import pytest
import seaborn
from matplotlib import pyplot

from wqio.features import Dataset, Location
from wqio.tests import helpers

BASELINE_IMAGES = "_baseline_images/features_tests"
TOLERANCE = helpers.get_img_tolerance()


@helpers.seed
def setup_location(station_type):
    data = helpers.getTestROSData()
    loc = Location(
        data,
        station_type=station_type,
        bsiter=10000,
        rescol="res",
        qualcol="qual",
        useros=True,
    )
    pyplot.rcdefaults()
    return loc


@pytest.fixture
def inflow_loc_blue():
    loc = setup_location("inflow")
    loc.color = "cornflowerblue"
    loc.plot_marker = "o"
    return loc


@pytest.fixture
def inflow_loc_red():
    loc = setup_location("inflow")
    loc.color = "firebrick"
    loc.plot_marker = "d"
    return loc


@pytest.fixture
def inflow_loc_green():
    loc = setup_location("inflow")
    loc.color = "forestgreen"
    loc.plot_marker = "d"
    return loc


@pytest.fixture
def inflow_xlims():
    return {"left": 0, "right": 2}


@pytest.fixture
def inflow_xlims_vs():
    return {"left": -0.5, "right": 0.5}


@pytest.fixture
def outflow_loc():
    loc = setup_location("outflow")
    return loc


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_loc_boxplot_default(inflow_loc_blue, inflow_xlims):
    fig1 = inflow_loc_blue.boxplot()
    return fig1


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_loc_boxplot_patch_artists(inflow_loc_blue, inflow_xlims):
    fig2 = inflow_loc_blue.boxplot(patch_artist=True, xlims=inflow_xlims)
    return fig2


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_loc_boxplot_linscale(inflow_loc_blue, inflow_xlims):
    fig3 = inflow_loc_blue.boxplot(yscale="linear", xlims=inflow_xlims)
    return fig3


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_loc_boxplot_no_mean(inflow_loc_red, inflow_xlims):
    fig4 = inflow_loc_red.boxplot(showmean=False, xlims=inflow_xlims)
    return fig4


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_loc_boxplot_width(inflow_loc_red, inflow_xlims):
    fig5 = inflow_loc_red.boxplot(width=1.25, xlims=inflow_xlims)
    return fig5


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_loc_boxplot_no_notch(inflow_loc_red, inflow_xlims):
    fig6 = inflow_loc_red.boxplot(shownotches=False, xlims=inflow_xlims)
    return fig6


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_loc_boxplot_bacteria_geomean(inflow_loc_green, inflow_xlims):
    fig7 = inflow_loc_green.boxplot(bacteria=True, xlims=inflow_xlims)
    return fig7


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_loc_boxplot_with_ylabel(inflow_loc_green, inflow_xlims):
    fig8 = inflow_loc_green.boxplot(ylabel="Test Ylabel", xlims=inflow_xlims)
    return fig8


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_loc_boxplot_provided_ax(inflow_loc_green, inflow_xlims):
    fig10, ax10 = pyplot.subplots()
    fig10 = inflow_loc_green.boxplot(ax=ax10, xlims=inflow_xlims)
    assert isinstance(fig10, pyplot.Figure)
    return fig10


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_loc_boxplot_custom_position(inflow_loc_green, inflow_xlims):
    fig11 = inflow_loc_green.boxplot(pos=1.5, xlims=inflow_xlims)
    return fig11


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_loc_boxplot_with_xlabel(inflow_loc_green, inflow_xlims):
    fig12 = inflow_loc_green.boxplot(xlabel="Test Xlabel", xlims=inflow_xlims)
    return fig12


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_loc_probplot_default(inflow_loc_blue):
    fig1 = inflow_loc_blue.probplot()
    return fig1


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_loc_probplot_provided_ax(inflow_loc_blue):
    fig2, ax2 = pyplot.subplots()
    fig2 = inflow_loc_blue.probplot(ax=ax2)
    assert isinstance(fig2, pyplot.Figure)
    return fig2


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_loc_probplot_yscale_linear(inflow_loc_blue):
    fig3 = inflow_loc_blue.probplot(yscale="linear")
    return fig3


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_loc_probplot_ppax(inflow_loc_blue):
    fig4 = inflow_loc_blue.probplot(axtype="pp")
    return fig4


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_loc_probplot_qqax(inflow_loc_blue):
    fig5 = inflow_loc_blue.probplot(axtype="qq")
    return fig5


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_loc_probplot_ylabel(inflow_loc_red):
    fig6 = inflow_loc_red.probplot(ylabel="test ylabel")
    return fig6


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_loc_probplot_clear_yticks(inflow_loc_red):
    fig7 = inflow_loc_red.probplot(clearYLabels=True)
    return fig7


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_loc_probplot(inflow_loc_red):
    fig8 = inflow_loc_red.probplot()
    return fig8


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_loc_probplot_no_rotate_xticklabels(inflow_loc_green):
    fig10 = inflow_loc_green.probplot(rotateticklabels=False)
    return fig10


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_loc_probplot_plotopts1(inflow_loc_green):
    fig12 = inflow_loc_green.probplot(
        markersize=10,
        linestyle="--",
        color="blue",
        markerfacecolor="none",
        markeredgecolor="green",
    )
    return fig12


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_loc_probplot_plotopts2(inflow_loc_green):
    fig13 = inflow_loc_green.probplot(
        markeredgewidth=2, markerfacecolor="none", markeredgecolor="green"
    )
    return fig13


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_loc_statplot_custom_position(inflow_loc_blue):
    fig1 = inflow_loc_blue.statplot(pos=1.25)
    return fig1


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_loc_statplot_yscale_linear(inflow_loc_blue):
    fig2 = inflow_loc_blue.statplot(yscale="linear")
    return fig2


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_loc_statplot_no_notch(inflow_loc_blue):
    fig3 = inflow_loc_blue.statplot(shownotches=False)
    return fig3


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_loc_statplot_no_mean(inflow_loc_red):
    fig4 = inflow_loc_red.statplot(showmean=False)
    return fig4


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_loc_statplot_custom_width(inflow_loc_red):
    fig5 = inflow_loc_red.statplot(width=1.5)
    return fig5


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_loc_statplot_bacteria_true(inflow_loc_red):
    fig6 = inflow_loc_red.statplot(bacteria=True)
    return fig6


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_loc_statplot_ylabeled(inflow_loc_red):
    fig7 = inflow_loc_red.statplot(ylabel="Test Y-Label")
    return fig7


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_loc_statplot_qq(inflow_loc_green):
    fig8 = inflow_loc_green.statplot(axtype="qq")
    return fig8


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_loc_statplot_pp(inflow_loc_green):
    fig9 = inflow_loc_green.statplot(axtype="pp")
    return fig9


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_loc_statplot_patch_artist(inflow_loc_green):
    fig10 = inflow_loc_green.statplot(patch_artist=True)
    assert isinstance(fig10, pyplot.Figure)
    return fig10


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_loc_vertical_scatter_default(inflow_loc_blue):
    fig = inflow_loc_blue.verticalScatter()
    return fig


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_loc_vertical_scatter_provided_ax(inflow_loc_blue):
    fig, ax2 = pyplot.subplots()
    fig = inflow_loc_blue.verticalScatter(ax=ax2)
    assert isinstance(fig, pyplot.Figure)
    return fig


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_loc_vertical_scatter_pos(inflow_loc_blue):
    fig = inflow_loc_blue.verticalScatter(pos=1.25)
    return fig


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_loc_vertical_scatter_ylabel(inflow_loc_red):
    fig = inflow_loc_red.verticalScatter(ylabel="Test Y-Label")
    return fig


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_loc_vertical_scatter_yscale_linear(inflow_loc_red):
    fig = inflow_loc_red.verticalScatter(yscale="linear")
    return fig


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_loc_vertical_scatter_not_ignoreROS(inflow_loc_red):
    fig = inflow_loc_red.verticalScatter(ignoreROS=False)
    return fig


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_loc_vertical_scatter_markersize(inflow_loc_red):
    fig = inflow_loc_red.verticalScatter(markersize=8)
    return fig


@helpers.seed
def setup_dataset(extra_NDs=False):
    in_data = helpers.getTestROSData()
    in_data["res"] += 3

    out_data = helpers.getTestROSData()
    out_data["res"] -= 1.5

    if extra_NDs:
        in_data.loc[[0, 1, 2], "qual"] = "ND"
        out_data.loc[[14, 15, 16], "qual"] = "ND"

    influent = Location(
        in_data,
        station_type="inflow",
        bsiter=10000,
        rescol="res",
        qualcol="qual",
        useros=False,
    )

    effluent = Location(
        out_data,
        station_type="outflow",
        bsiter=10000,
        rescol="res",
        qualcol="qual",
        useros=False,
    )

    ds = Dataset(influent, effluent, name="Test Dataset")
    pyplot.rcdefaults()
    return ds


@pytest.fixture
def ds_basic():
    return setup_dataset(extra_NDs=False)


@pytest.fixture
def ds_NDs():
    return setup_dataset(extra_NDs=True)


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_ds_boxplot_default(ds_basic, inflow_xlims):
    fig1 = ds_basic.boxplot()
    return fig1


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_ds_boxplot_patch_artists(ds_basic, inflow_xlims):
    fig2 = ds_basic.boxplot(patch_artist=True, xlims=inflow_xlims)
    return fig2


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_ds_boxplot_linscale(ds_basic, inflow_xlims):
    fig3 = ds_basic.boxplot(yscale="linear", xlims=inflow_xlims)
    return fig3


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_ds_boxplot_no_mean(ds_basic, inflow_xlims):
    fig4 = ds_basic.boxplot(showmean=False, xlims=inflow_xlims)
    return fig4


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_ds_boxplot_width(ds_basic, inflow_xlims):
    fig5 = ds_basic.boxplot(width=1.25, xlims=inflow_xlims)
    return fig5


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_ds_boxplot_no_notch(ds_basic, inflow_xlims):
    fig6 = ds_basic.boxplot(shownotches=False, xlims=inflow_xlims)
    return fig6


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_ds_boxplot_bacteria_geomean(ds_basic, inflow_xlims):
    fig7 = ds_basic.boxplot(bacteria=True, xlims=inflow_xlims)
    return fig7


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_ds_boxplot_with_ylabel(ds_basic, inflow_xlims):
    fig8 = ds_basic.boxplot(ylabel="Test Ylabel", xlims=inflow_xlims)
    return fig8


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_ds_boxplot_provided_ax(ds_basic, inflow_xlims):
    fig10, ax10 = pyplot.subplots()
    fig10 = ds_basic.boxplot(ax=ax10, xlims=inflow_xlims)
    assert isinstance(fig10, pyplot.Figure)

    return fig10


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_ds_boxplot_custom_position(ds_basic, inflow_xlims):
    fig11 = ds_basic.boxplot(pos=1.5, xlims=inflow_xlims)
    return fig11


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_ds_boxplot_custom_offset(ds_basic, inflow_xlims):
    fig12 = ds_basic.boxplot(offset=0.75, xlims=inflow_xlims)
    return fig12


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_ds_boxplot_single_tick(ds_basic, inflow_xlims):
    fig13 = ds_basic.boxplot(bothTicks=False, xlims=inflow_xlims)
    return fig13


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_ds_boxplot_single_tick_no_name(ds_basic, inflow_xlims):
    ds_basic.name = None
    fig14 = ds_basic.boxplot(bothTicks=False, xlims=inflow_xlims)
    return fig14


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_ds_probplot_default(ds_basic):
    fig1 = ds_basic.probplot()
    return fig1


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_ds_probplot_provided_ax(ds_basic):
    fig2, ax2 = pyplot.subplots()
    fig2 = ds_basic.probplot(ax=ax2)
    assert isinstance(fig2, pyplot.Figure)
    return fig2


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_ds_probplot_yscale_linear(ds_basic):
    fig3 = ds_basic.probplot(yscale="linear")
    return fig3


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_ds_probplot_ppax(ds_basic):
    fig4 = ds_basic.probplot(axtype="pp")
    return fig4


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_ds_probplot_qqax(ds_basic):
    fig5 = ds_basic.probplot(axtype="qq")
    return fig5


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_ds_probplot_ylabel(ds_basic):
    fig6 = ds_basic.probplot(ylabel="test ylabel")
    return fig6


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_ds_probplot_clear_yticks(ds_basic):
    fig7 = ds_basic.probplot(clearYLabels=True)
    return fig7


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_ds_probplot(ds_basic):
    fig8 = ds_basic.probplot()
    return fig8


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_ds_probplot_no_rotate_xticklabels(ds_basic):
    fig10 = ds_basic.probplot(rotateticklabels=False)
    return fig10


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_ds_statplot_custom_position(ds_basic):
    fig1 = ds_basic.statplot(pos=1.25)
    return fig1


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_ds_statplot_yscale_linear(ds_basic):
    fig2 = ds_basic.statplot(yscale="linear")
    return fig2


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_ds_statplot_no_notch(ds_basic):
    fig3 = ds_basic.statplot(shownotches=False)
    return fig3


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_ds_statplot_no_mean(ds_basic):
    fig4 = ds_basic.statplot(showmean=False)
    return fig4


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_ds_statplot_custom_width(ds_basic):
    fig5 = ds_basic.statplot(width=1.5)
    return fig5


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_ds_statplot_bacteria_true(ds_basic):
    fig6 = ds_basic.statplot(bacteria=True)
    return fig6


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_ds_statplot_ylabeled(ds_basic):
    fig7 = ds_basic.statplot(ylabel="Test Y-Label")
    return fig7


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_ds_statplot_qq(ds_basic):
    fig8 = ds_basic.statplot(axtype="qq")
    return fig8


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_ds_statplot_pp(ds_basic):
    fig9 = ds_basic.statplot(axtype="pp")
    return fig9


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_ds_statplot_patch_artist(ds_basic):
    fig10 = ds_basic.statplot(patch_artist=True)
    assert fig10, pyplot.Figure
    return fig10


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_ds_scatterplot_default(ds_NDs):
    fig1 = ds_NDs.scatterplot()
    return fig1


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_ds_scatterplot_best_fit(ds_NDs):
    fig = ds_NDs.scatterplot(bestfit=True)
    return fig


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_ds_scatterplot_best_fit_through_origin(ds_NDs):
    with numpy.errstate(divide="ignore", invalid="ignore"):
        fig = ds_NDs.scatterplot(bestfit=True, fitopts=dict(through_origin=True))
    return fig


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_ds_scatterplot_provided_ax(ds_NDs):
    fig2, ax2 = pyplot.subplots()
    fig2 = ds_NDs.scatterplot(ax=ax2)
    assert isinstance(fig2, pyplot.Figure)

    return fig2


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_ds_scatterplot_xscale_linear(ds_NDs):
    fig3 = ds_NDs.scatterplot(xscale="linear")
    return fig3


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_ds_scatterplot_xyscale_linear(ds_NDs):
    fig5 = ds_NDs.scatterplot(xscale="linear", yscale="linear")
    return fig5


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_ds_scatterplot_yscale_linear(ds_NDs):
    fig4 = ds_NDs.scatterplot(yscale="linear")
    return fig4


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_ds_scatterplot_xlabel(ds_NDs):
    fig6 = ds_NDs.scatterplot(xlabel="X-label")
    return fig6


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_ds_scatterplot_ylabel(ds_NDs):
    fig7 = ds_NDs.scatterplot(ylabel="Y-label")
    return fig7


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_ds_scatterplot_no_xlabel(ds_NDs):
    fig8 = ds_NDs.scatterplot(xlabel="")
    return fig8


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_ds_scatterplot_no_ylabel(ds_NDs):
    fig9 = ds_NDs.scatterplot(ylabel="")
    return fig9


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_ds_scatterplot_no_legend(ds_NDs):
    fig10 = ds_NDs.scatterplot(showlegend=False)
    return fig10


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_ds_scatterplot_one2one(ds_NDs):
    fig10 = ds_NDs.scatterplot(one2one=True)
    return fig10


def test_ds_scatterplot_useros(ds_NDs):
    with helpers.raises(ValueError):
        ds_NDs.scatterplot(useros=True)


@pytest.fixture
def markerkwargs():
    opts = dict(
        linestyle="none",
        markerfacecolor="black",
        markeredgecolor="white",
        markeredgewidth=0.5,
        markersize=6,
        zorder=10,
    )
    return opts


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_ds__plot_NDs_both(ds_NDs, markerkwargs):
    fig1, ax1 = pyplot.subplots()
    ds_NDs._plot_nds(ax1, which="both", marker="d", **markerkwargs)
    ax1.set_xlim(0, 30)
    ax1.set_ylim(0, 30)
    return fig1


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_ds__plot_NDs_effluent(ds_NDs, markerkwargs):
    fig2, ax2 = pyplot.subplots()
    ds_NDs._plot_nds(ax2, which="effluent", marker="<", **markerkwargs)
    ax2.set_xlim(0, 30)
    ax2.set_ylim(0, 30)
    return fig2


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_ds__plot_NDs_influent(ds_NDs, markerkwargs):
    fig3, ax3 = pyplot.subplots()
    ds_NDs._plot_nds(ax3, which="influent", marker="v", **markerkwargs)
    ax3.set_xlim(0, 30)
    ax3.set_ylim(0, 30)
    return fig3


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_ds__plot_NDs_neither(ds_NDs, markerkwargs):
    fig4, ax4 = pyplot.subplots()
    ds_NDs._plot_nds(ax4, which="neither", marker="o", **markerkwargs)
    ax4.set_xlim(0, 30)
    ax4.set_ylim(0, 30)
    return fig4


def _do_jointplots(ds, hist=False, kde=False, rug=False):
    jg = ds.jointplot(hist=hist, kde=kde, rug=rug)
    assert isinstance(jg, seaborn.JointGrid)
    return jg.figure


def test_ds_joint_hist_smoke(ds_NDs):
    with seaborn.axes_style("ticks"):
        fig1 = _do_jointplots(ds_NDs, hist=True)
    return fig1


def test_ds_joint_kde_smoke(ds_NDs):
    with seaborn.axes_style("ticks"):
        fig2 = _do_jointplots(ds_NDs, kde=True)
    return fig2


def test_ds_joint_rug_smoke(ds_NDs):
    with seaborn.axes_style("ticks"):
        fig3 = _do_jointplots(ds_NDs, rug=True)
        return fig3


def test_ds_joint_kde_rug_hist_smoke(ds_NDs):
    with seaborn.axes_style("ticks"):
        fig4 = _do_jointplots(ds_NDs, hist=True, kde=True)
    return fig4
